/*
** config.c - read a config file & process it - part of GrabURL 2.0
** Copyright (C) 1996-97 Serge Emond
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "libgurl.h"
#include "config.h"
#include "graburl.h"
#include "Base64.h"
#include "macros.h"
#include "uri.h"

#include "license.h" /* Help text and license */

struct Cfg_Global cfg_global;

struct TmpList *urilist, *urilistp, *uriflist, *uriflistp;

/*
** Initialize private stuff for each "section"
*/

void InitConfig()
{
  int i;
  
  for (i=0; cfg_init[i].cmd; i++)
    cfg_init[i].cmd();
}

/*
** Open the config-file cfgname, and do what is to be done...
*/

void ReadConfigFile(char *cfgname)
{
  FILE *cfh;				/* Config file handler */
  char line[MAX_CFG_LINE_LEN], *p, *section=0, *cret;
  int i, linen=0, match;
  
  cfh = fopen(cfgname, "r");
  if (!cfh) {
    perror("ReadConfigFile");
    exit(EXIT_FAILURE);
  }
  
  while (fsgetdelim(cfh, line, MAX_CFG_LINE_LEN, 0, "\n") > 0) {
    linen++;
    p = NextNonBlank(line);
    if (!p) continue;
    /* Comments */
    if (p[0]=='#') continue;
    /* Special commands */
    
    if (!section && !str_nieq(p, "section", 7)) {
      int secok;
      
      if (section) {
	fprintf(stderr, "Error in config file \"%s\" line %d:"
		" Cannot use 'section' in a section\n",
		cfgname, linen);
	exit(EXIT_FAILURE);
      }
      
      section = ExtractArgument(p, 1);
      if (!section) {
	fprintf(stderr, "Error in config file \"%s\" line %d:"
		" Missing argument to 'section'\n",
		cfgname, linen);
	exit(EXIT_FAILURE);
      }
      
      secok = 0;
      for (i=0; cfg_sections[i].name; i++) {
	if (!str_ieq(section, cfg_sections[i].name)) {
	  secok=1;
	  break;
	}
      }
      if (!secok) {
	fprintf(stderr, "Error in config file \"%s\" line %d:"
		" Section \"%s\" unknown\n",
		cfgname, linen, section);
	exit(EXIT_FAILURE);
      }
    }
    else if (!str_nieq(p, "end", 3)) {
      if (!section) {
	fprintf(stderr, "Error in config file \"%s\" line %d:"
		" 'end' not in a section\n",
		cfgname, linen);
	exit(EXIT_FAILURE);
      }
      free(section);
      section=0;
    }
    else {
      match = 0;
      for (i=0; cfg_linecmds[i].tag; i++) {
	if ( (!section && !cfg_linecmds[i].section)
	     || (section && cfg_linecmds[i].section
		 && !str_ieq(section, cfg_linecmds[i].section)) ) {
	  if (!str_nieq(p, cfg_linecmds[i].tag, cfg_linecmds[i].tagl)
	      && cfg_linecmds[i].cmd) {
	    cret = cfg_linecmds[i].cmd(p);
	    if (cret) {
	      fprintf(stderr, "Error in config file \"%s\" line %d: %s\n",
		      cfgname, linen, cret);
	      exit(EXIT_FAILURE);
	    }
	    match = 1;
	    break;
	  }
	}
      }
      
      if (!match) {
	fprintf(stderr, "Error in config file \"%s\" line %d:"
		" Unknown command\n",
		cfgname, linen);
	exit(EXIT_FAILURE);
      }
    }
  } /* while */
  
  fclose(cfh);
}

char *cfg_cmd_include(const char *line)
{
  char *arg;
  
  arg = ExtractArgument(line, 1);
  if (!arg) return "missing argument to 'include'";
  
  ReadConfigFile(arg);
  free(arg);
  return 0;
}

char *cfg_cmd_email(const char *line)
{
  ;

  if (cfg_global.fromstring) return "Second email definition ignored";
  cfg_global.fromstring = ExtractArgument(line, 1);
  if (!cfg_global.fromstring) return "Missing argument to EMail";

  return 0;
}

char *cfg_cmd_delay(const char *line)
{
  char *arg;
  int del;
  
  arg = ExtractArgument(line, 1);
  if (!arg) return "Missing argument to 'delay'";
  
  del = strtol(arg, 0, 10);
  free(arg);
  cfg_global.wait.tv_sec = del / 10;
  cfg_global.wait.tv_usec = (del % 10) * 100000;
  return 0;
}

char *cfg_cmd_saveroot(const char *line)
{
  ;
  
  if (cfg_global.saveroot) return "Warning: SaveRoot already defined";
  
  cfg_global.saveroot = ExtractArgument(line, 1);
  if (!cfg_global.saveroot) return "Missing argument to SaveRoot";
  return 0;
}

char *cfg_cmd_dirmode(const char *line)
{
  char *arg;
  
  arg = ExtractArgument(line, 1);
  if (!arg) return "Missing argument to DirMode";
  
  cfg_global.d_mode = strtol(arg, 0, 8);
  free(arg);
  return 0;
}

char *cfg_cmd_filemode(const char *line)
{
  char *arg;
  
  arg = ExtractArgument(line, 1);
  if (!arg) return "Missing argument to FileMode";
  
  cfg_global.f_mode = strtol(arg, 0, 8);
  free(arg);
  return 0;
}

char *cfg_cmd_translate(const char *line)
{
  if (cfg_global.trans_from || cfg_global.trans_to)
    return "Translate already defined";
  
  cfg_global.trans_from = ExtractArgument(line, 1);
  cfg_global.trans_to = ExtractArgument(line, 2);
  if (!cfg_global.trans_from || !cfg_global.trans_to) {
    if (cfg_global.trans_from) {
      free(cfg_global.trans_from);
      cfg_global.trans_from = 0;
    }
    if (cfg_global.trans_to) {
      free(cfg_global.trans_to);
      cfg_global.trans_to = 0;
    }
    return "Missing argument(s) to Translate";
  }
  return 0;
}

char *cfg_cmd_retry(const char *line)
{
  cfg_global.def_flags |= URI_RETRY;
  return 0;
}

char *cfg_cmd_scan(const char *line)
{
  cfg_global.def_flags |= URI_SCAN;
  return 0;
}

char *cfg_cmd_stayinhost(const char *line)
{
  cfg_global.def_flags |= URI_STAYINHOST;
  return 0;
}

char *cfg_cmd_nonexists(const char *line)
{
  cfg_global.def_flags |= URI_NONEXISTS;
  return 0;
}

/* The following ones have to be in section "http" */

char *cfg_cmd_auth(const char *line)
{
  char *arg, *arg2;
  struct Cfg_Auth *cauth;
  
  arg = ExtractArgument(line, 1);
  if (arg) {
    arg2 = ExtractArgument(line, 2);
    if (!arg2) {
      free(arg); arg = 0;
    }
  }
  
  if (!arg) return "Missing argument(s) to Auth";
  
  cauth = (struct Cfg_Auth *)gmalloc(sizeof(struct Cfg_Auth));
  cauth->realm = arg;
  cauth->upw = EnBaseString(arg2);
  free(arg2);
  cauth->next = cfg_global.auth;
  cfg_global.auth = cauth;
  return 0;
}

char *cfg_cmd_requestline(const char *line)
{
  struct RequestLine *rl;
  char *arg;
  
  arg = ExtractArgument(line, 1);
  if (!arg) return "Missing argument to RequestLine";
  
  rl = (struct RequestLine *)gmalloc(sizeof (struct RequestLine));
  rl->next = cfg_global.shead;
  rl->line = arg;
  cfg_global.shead = rl;
  return 0;
}

char *cfg_cmd_depth(const char *line)
{
  char *arg;
  
  arg = ExtractArgument(line, 1);
  if (!arg) return "Missing argument to Depth";
  
  cfg_global.def_depth = atoi(arg);
  return 0;
}

/* Not implemented
  char *cfg_cmd_addhead(const char *line)
  {
  cfg_global.def_flags |= URI_ADDHEAD;
  return 0;
  }
*/

char *cfg_cmd_ifmodified(const char *line)
{
  cfg_global.def_flags |= URI_IFMODIFIED;
  return 0;
}

char *cfg_cmd_savehead(const char *line)
{
  cfg_global.def_flags |= URI_SAVEHEAD;
  return 0;
}

/* Command line arguments */

int GetConfigArg(char **argv)
{
  int i;
  
  for (i=0; argv[i]; i++) {
    if ((str_ieq(argv[i], "c")) == 0 || (str_ieq(argv[i], "config") == 0))
      return(i+1);
  }
  return(0);
}

void GetArgs(char **argv)
{
  int i, pos, delta;
  struct TmpList *p;
  
  urilist = 0; urilistp = 0; uriflist = 0; uriflistp = 0;
  
  for (pos=1; argv[pos]; pos++) {
    for (i=0, delta=0; cfg_args[i].cmd; i++) {
      delta = cfg_args[i].cmd(argv, pos);
      if (delta != 0) break;
    }
    
    if (delta > 0) pos += (delta-1);
    else {
      /* No arg found - we assume it is an URI */
      p = (struct TmpList *)gmalloc(sizeof(struct TmpList));
      p->str = (char *)gmalloc(strlen(argv[pos])+1);
      strcpy(p->str, argv[pos]);
      p->prev = urilistp;
      p->next = 0;
      if (urilistp) urilistp->next = p;
      urilistp = p;
      if (!urilist) urilist = urilistp;
    }
  }
}

void cfg_global_init(void)
{
  ;
  
  memset(&cfg_global, 0, sizeof(struct Cfg_Global));
  cfg_global.def_depth = -1;
}

void cfg_global_post(void)
{
  struct TmpList *tp, *tp2;
  
  grab = (struct Grab *)gmalloc(sizeof(struct Grab));
  
  memset(grab, 0, sizeof(struct Grab));
  
  if (!cfg_global.def_filename) {
    cfg_global.def_filename = (char *)gmalloc(11);
    strcpy(cfg_global.def_filename, "index.html");
  }
  
  if (!cfg_global.saveroot) {
    cfg_global.saveroot = (char *)gmalloc(2);
    strcpy(cfg_global.saveroot, ".");
  }
  
  if (!cfg_global.d_mode) cfg_global.d_mode = 0700;
  if (!cfg_global.f_mode) cfg_global.f_mode = 0700;
  
  /* And at the end.. we add uris and urifiles */
  for (tp = urilist; tp; tp = tp2) {
    if (!AddURI(0, tp->str, -2, -2, 0)) {
      fprintf(stderr, "Error adding uri \"%s\"\n", tp->str);
      exit(EXIT_FAILURE);
    }
    tp2 = tp->next;
    free(tp->str);
    free(tp);
  }
  
  for (tp = uriflist; tp; tp = tp2) {
    LoadURIFile(tp->str);
    tp2 = tp->next;
    free(tp);
  }
  
  if (cfg_global.def_saveto &&
      ((cfg_global.def_flags & URI_SCAN) || (CountURI(-1, -1) > 1))) {
    fprintf(stderr, "Cannot use global saveto with multiple"
	    " URLs or workfile\n");
    exit(EXIT_FAILURE);
  }
  
}


int rargs_global(char **argv, int pos)
{
  int delta;
  
  if ((str_ieq(argv[pos], "c") == 0) ||
      (str_ieq(argv[pos], "config") == 0))
    /* We ignore "config" argument */
    delta = 2;
  else if ((str_ieq(argv[pos], "?") == 0) ||
	   (str_ieq(argv[pos], "help") == 0) ||
	   (str_ieq(argv[pos], "--help") == 0) ||
	   (str_ieq(argv[pos], "-h") == 0)) {
    /* Show help, then exit */
    printf(HELP);
    exit(0);
  }
  else if (str_ieq(argv[pos], "license") == 0) {
    /* Show license then exit */
    printf(LICENSE);
    exit(0);
  }
  else if ((str_ieq(argv[pos], "sr") == 0) ||
	   (str_ieq(argv[pos], "saveroot") == 0)) {
    /* SaveRoot */
    if (cfg_global.saveroot) {
      free(cfg_global.saveroot);
    }
    
    cfg_global.saveroot = (char *)gmalloc(strlen(argv[++pos]+1));
    strcpy(cfg_global.saveroot, argv[pos]);
    delta = 2;
  }
  else if ((str_ieq(argv[pos], "if") == 0) ||
	   (str_ieq(argv[pos], "infile") == 0)) {
    /* Add filename to queue */
    struct TmpList *p;
    p = (struct TmpList *)gmalloc(sizeof(struct TmpList));
    p->str = argv[pos+1];
    p->prev = uriflistp;
    p->next = 0;
    uriflistp = p;
    if (p->prev) p->prev->next = p;
    if (!uriflist) uriflist = p;
    delta = 2;
  }
  else if (str_ieq(argv[pos], "retry") == 0) {
    cfg_global.def_flags |= URI_RETRY;
    delta = 1;
  }
  else if (str_ieq(argv[pos], "noretry") == 0) {
    cfg_global.def_flags &= ~URI_RETRY;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "r") == 0) ||
	   (str_ieq(argv[pos], "recursive") == 0) ||
	   (str_ieq(argv[pos], "scan") == 0)) {
    cfg_global.def_flags |= URI_SCAN;
    delta = 1;
  }
  else if (str_ieq(argv[pos], "noscan") == 0) {
    cfg_global.def_flags &= ~URI_SCAN;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "sih") == 0) ||
	   (str_ieq(argv[pos], "stayinhost") == 0)) {
    cfg_global.def_flags |= URI_STAYINHOST;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "nosih") == 0) ||
	   (str_ieq(argv[pos], "nostayinhost") == 0)) {
    cfg_global.def_flags &= ~URI_STAYINHOST;
    delta = 1;
  }
  /* Not implemented
  else if ((str_ieq(argv[pos], "ah") == 0) ||
    (str_ieq(argv[pos], "addhead") == 0)) {
    cfg_global.def_flags |= URI_ADDHEAD;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "noah") == 0) ||
	   (str_ieq(argv[pos], "noaddhead") == 0)) {
    cfg_global.def_flags &= ~URI_ADDHEAD;
    delta = 1;
  }
  */
  else if ((str_ieq(argv[pos], "im") == 0) ||
	   (str_ieq(argv[pos], "ifmodified") == 0)) {
    cfg_global.def_flags |= URI_IFMODIFIED;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "noim") == 0) ||
	   (str_ieq(argv[pos], "noifmodified") == 0)) {
    cfg_global.def_flags &= ~URI_IFMODIFIED;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "d") == 0) ||
	   (str_ieq(argv[pos], "depth") == 0)) {
    cfg_global.def_depth = strtol(argv[pos+1], 0, 10);
    delta = 2;
  }
  else if ((str_ieq(argv[pos], "ne") == 0) ||
	   (str_ieq(argv[pos], "nonexists") == 0)) {
    cfg_global.def_flags |= URI_NONEXISTS;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "none") == 0) ||
	   (str_ieq(argv[pos], "nononexists") == 0)) {
    cfg_global.def_flags &= ~URI_NONEXISTS;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "nd") == 0) ||
	   (str_ieq(argv[pos], "nodirs") == 0)) {
    cfg_global.def_flags |= URI_NODIRS;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "nond") == 0) ||
	   (str_ieq(argv[pos], "nonodirs") == 0)) {
    cfg_global.def_flags &= !URI_NODIRS;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "sh") == 0) ||
	   (str_ieq(argv[pos], "saveheader") == 0)) {
    cfg_global.def_flags |= URI_SAVEHEAD;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "nosh") == 0) ||
	   (str_ieq(argv[pos], "nosaveheader") == 0)) {
    cfg_global.def_flags &= ~URI_SAVEHEAD;
    delta = 1;
  }
  else if ((str_ieq(argv[pos], "wf") == 0) ||
	   (str_ieq(argv[pos], "workfile") == 0)) {
    if (cfg_global.workfile) {
      fprintf(stderr, "Cannot multiple workfiles\n");
      exit(EXIT_FAILURE);
    }
    cfg_global.workfile = (char *)gmalloc(strlen(argv[++pos]) + 1);
    strcpy(cfg_global.workfile, argv[pos]);
    delta = 2;
  }
  else if ((str_ieq(argv[pos], "st") == 0) ||
	   (str_ieq(argv[pos], "saveto") == 0)) {
    if (cfg_global.def_saveto) {
      fprintf(stderr, "Cannot use multiple global 'saveto'\n");
      exit(EXIT_FAILURE);
    }
    cfg_global.def_saveto = (char *)gmalloc(strlen(argv[++pos])+1);
    strcpy(cfg_global.def_saveto, argv[pos]);
    delta = 2;
  }
  else
    delta = 0;
  
  return(delta);
}

void PostConfig()
{
  int i;
  
  for (i=0; cfg_post[i].cmd; i++)
    cfg_post[i].cmd();
}
